from nonebot import on_command
from nonebot.adapters import Bot, Event
from nonebot.adapters.cqhttp import MessageSegment
from nonebot.log import logger
from nonebot.typing import T_State

from .ascii2d import Ascii2D
from .saucenao import SauceNao
from .utils import get_message_image
from .config import search_proxy, saucenao_api_key

logger.warning(f"PROXY: {search_proxy}")
logger.warning(f"S_API: {saucenao_api_key}")
Search = on_command('search')


def have_image(images: list) -> bool:
    return len(images) > 0

saucenao = SauceNao(saucenao_api_key, search_proxy)
ascii2d = Ascii2D(search_proxy)

@Search.handle()
async def search(bot: Bot, event: Event, state: T_State):
    images = get_message_image(event.json())
    if have_image(images):
        for image in images:
            logger.info(f"imgsearch: search -> \"{image}\"")

            logger.info(f"SauceNAO: searching...")
            res_sauce = await saucenao.search(image)
            if res_sauce.status_code//100 == 2:
                logger.info(f"SauceNAO: hit on {res_sauce.content['rate']}")
                res_text = f"[ {res_sauce.content['index']} / {res_sauce.content['rate']} ]\n" + \
                    "\n".join(f"{k}: {v}"for k,
                              v in res_sauce.content['data'].items())
                message = MessageSegment.reply(
                    event.message_id)+MessageSegment.text(res_text)
                await bot.send(event, message)
            elif res_sauce.status_code//100 == 3:
                logger.info("SauceNAO: not found")
                logger.info(f"Ascii2D: searching...")
                res_ascii = await ascii2d.search(image)
                if res_ascii.status_code//100 == 2:
                    message = MessageSegment.reply(event.message_id)+MessageSegment.text(
                        "\n".join(f"{k}: {v}"for k, v in res_ascii.content.items()))
                    await bot.send(event, message)

                elif res_ascii.status_code//100 == 3:
                    message_1 = MessageSegment.reply(event.message_id)+MessageSegment.text("\n".join(f"{k}: {v}"for k, v in res_ascii.content[0].items())) +\
                        MessageSegment.image(res_ascii.content[1])
                    message_2 = MessageSegment.reply(event.message_id)+MessageSegment.text("\n".join(f"{k}: {v}"for k, v in res_ascii.content[2].items())) +\
                        MessageSegment.image(res_ascii.content[3])
                    logger.info(f"Ascii2D: sending possible results...")
                    await bot.send(event, message_1)
                    await bot.send(event, message_2)
                elif res_ascii.status_code//100 == 4:
                    logger.info(f"Ascii2D: {res_ascii.message}")

            elif res_sauce.status_code//100 == 4:
                logger.error(f"SauceNAO: {res_sauce.message}")



from typing import Literal, Optional, Union
import json
from PIL import Image
from io import BytesIO


def get_message_image(data:Union[str,dict]) -> "list":
    """
    返回一个包含消息中所有图片文件路径的list,

    Args : 
          * ``data: str`` : 消息内容, 来自event.json()  
          * ``type: Literal['file','url']``: 当``type``为``'file'``时, 返回的是文件路径, 当``type``为``'url'``时, 返回的是url  

    Return :
          * ``img_list: list`` : 包含图片绝对路径或url的list
    """
    if isinstance(data,str):
        data = json.loads(data)
    return  [message['data']['url'] for message in data['message'] if message['type'] == 'image']


def compress_image(bytes_image: bytes) -> "bytes":
    """
    压缩图片, 搞小一点，excited!

    Args :
          * ``bytes_image: bytes``: 图片的bytes数据

    Return :
          * ``image: bytes``: 压缩后的图片
    """

    image = Image.open(BytesIO(bytes_image))
    image.thumbnail((250,250),resample=Image.ANTIALIAS)
    image_data = BytesIO()
    image.save(image_data, format='JPEG')
    return image_data.getvalue()




# -*- coding:utf-8 -*-
import json
import re
from collections import OrderedDict
from PIL import Image
from io import BytesIO
import httpx
from loguru import logger
from .response import *
from .utils import compress_image

Response = BaseResponse


class SauceNao:
    """ # SauceNAO search moudle #"""

    __instance = None

    def __init__(self, saucenao_api: str, proxy: None) -> None:
        self.log = ""  # 每次搜索后更新本次日志
        self.short_remaing = 0  # 每次搜索后更新30s剩余次数
        self.long_remaing = 0  # 每次搜索后更新24h剩余次数
        self.proxy = proxy  # 魔法
        self.raw = ""  # 每次搜索的原始数据
        self.saucenao_api_key = saucenao_api
        if not self.saucenao_api_key:
            raise Exception("没有设置saucenao_api_key，插件无法正常工作")

    def __new__(cls, *a, **k):
        if not cls.__instance:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    async def search(self, url: str) -> "Response":
        """ # SauceNAO search moudle #
        在SauceNAO中搜索这张图, 采用api模式而非网页解析, 所以你需要先去SauceNAO注册一个APIKEY

        :Args:
        * ``file_path:str`` target picture
        * ``APIKEY: str`` saucenao APIKEY (apply from https://saucenao.com/account/register)
        * ``proxy: str`` proxy (default: None , edit in .env file)

        :Returns:
        ``Response: Response``

        当搜索结果相似度小于75%时:
        >>> Response.status = ACTION_WARNING
        >>> Response.message:dict #此api在30s | 24h 内的剩余搜索次数
        >>> Response.message:str #相似度信息

        当搜索结果相似度大于75%时:
        >>> Response.status = ACTION_SUCCESS
        >>> Response.message:str #此api剩余搜索次数
        >>> Response.content:dict #返回搜索结果字典

        其他情况:
        >>> Response.status = ACTION_FAILED
        >>> Response.message:str #错误信息
        """
        bitmask_all = '999'  # 搜索saucenao全部index
        default_minsim = '75!'  # 最小匹配相似度
        #ImageFile.LOAD_TRUNCATED_IMAGES = True  # qq有时候拿到的是烂图, 不完整的
        client = httpx.AsyncClient(proxies=None)
        
        
        for i in range(3):
            try:
                img = (await client.get(url)).content
                await client.aclose()
                break
            except httpx.ReadTimeout:
                logger.warning(f"SauceNAO: 下载qq图片超时, 尝试次数 {i+1}/3")
        else:
            await client.aclose()
            return Response(ACTION_FAILED,message="图片下载失败, 请检查网络是否通畅.")
            
        
        url_all = 'http://saucenao.com/search.php?output_type=2&numres=1&minsim=' + default_minsim + '&db=' + bitmask_all + '&api_key=' + self.saucenao_api_key
        files = {'file': ('img.jpeg', compress_image(img),'image/jpeg')}
        r = None

        for i in range(3):
            try:
                async with httpx.AsyncClient(proxies=self.proxy) as client:
                    r = await client.post(url=url_all, files=files,follow_redirects=True)
                break
            except httpx.ProxyError as e:
                logger.warning(f"SauceNAO: 代理错误, 尝试次数 {i+1}/3 {e}")
            except httpx.ReadTimeout as e:
                logger.warning(f"SauceNAO: 下载图片超时, 尝试次数 {i+1}/3 {e}")
            except httpx.ConnectError as e:
                logger.warning(f"SauceNAO: 连接错误, 请检查网络. 尝试次数 {i+1}/3 {e}")
        else:
            return Response(ACTION_FAILED,message="向SauceNAO请求失败")

        if r.status_code != 200:
            if r.status_code == 403:
                return Response(ACTION_FAILED, message="SAUCENAO_APIKEY 设置错误")
            else:
                return Response(ACTION_FAILED, message="其他错误, Error Code: " + str(r.status_code))
        else:
            results = json.JSONDecoder(
                object_pairs_hook=OrderedDict).decode(r.text)
            if int(results['header']['user_id']) > 0:
                _remain_searches = 'Remaining Searches 30s|24h: ' + \
                    str(results['header']['short_remaining']) + \
                    '|' + str(results['header']['long_remaining'])
                # print(_remain_searches)
                self.short_remaing = results['header']['short_remaining']
                self.long_remaing = results['header']['long_remaining']
                if int(results['header']['status']) == 0:
                    ...
                else:
                    if int(results['header']['status']) > 0:
                        return Response(ACTION_FAILED, message="SauceNAO服务器部分索引出现问题, 请稍后再试")
                    else:
                        return Response(ACTION_FAILED, message="图片无法读取或其他请求错误")
            else:
                return Response(ACTION_FAILED, message="图片无法读取或API错误")

            found_json = {'index': "", 'rate': "", 'data': {}}
            if int(results['header']['results_returned']) > 0:
                artwork_url = ""
                # print(results)
                self.raw = results
                rate = results['results'][0]['header']['similarity']+'%'
                # one or more results were returned
                if float(results['results'][0]['header']['similarity']) > float(results['header']['minimum_similarity']):
                    #print('hit! ' + str(results['results'][0]['header']['similarity']))

                    illust_id = 0
                    member_id = -1
                    index_id = results['results'][0]['header']['index_id']
                    page_string = ''
                    page_match = re.search(
                        '(_p[\d]+)\.', results['results'][0]['header']['thumbnail'])
                    if page_match:
                        page_string = page_match.group(1)

                    _data = results['results'][0]['data']
                    """
                    这里包含了绝大部分经常出现的搜索结果索引index, 
                    部分index由copilot自动补全,
                    如果是没包含到的index, 会返回给bot Unhandeled Index, 本次搜索结果的原始数据在self.raw中
                    """
                    if index_id == 5 or index_id == 6:
                        # 5->pixiv 6->pixiv historical
                        found_json['index'] = "pixiv"
                        member_name = _data['member_name']
                        illust_id = _data['pixiv_id']
                        title = _data['title']
                        artwork_url = f"https://pixiv.net/artworks/{illust_id}"
                        found_json['data'] = {
                            "title": title, "illust": illust_id, "author": member_name, "artwork": artwork_url}
                    elif index_id == 8:
                        # 8->nico nico seiga
                        found_json['index'] = 'seiga'
                        member_id = _data['member_id']
                        illust_id = _data['seiga_id']
                        found_json['data'] = {
                            "author": member_id, "illust": illust_id}
                    elif index_id == 9:
                        # 9 -> danbooru
                        # index name, danbooru_id, gelbooru_id, creator, material, characters, sources
                        found_json['index'] = "danbooru"
                        creator = _data['creator']
                        characters = _data['characters']
                        source = _data['source']
                        found_json['data'] = {
                            "creator": creator, "characters": characters, "source": source}
                    elif index_id == 10:
                        # 10->drawr
                        found_json['index'] = 'drawr'
                        member_id = _data['member_id']
                        illust_id = _data['drawr_id']
                        found_json['data'] = {
                            "member_id": member_id, "illust_id": illust_id}
                    elif index_id == 11:
                        # 11->nijie
                        found_json['index'] = 'nijie'
                        member_id = _data['member_id']
                        illust_id = _data['nijie_id']
                        found_json['data'] = {
                            "member_id": member_id, "illust_id": illust_id}
                    elif index_id == 12:
                        # 12 -> Yande.re
                        # ext_urls, yandere_id, creator, material, characters, source
                        found_json['index'] = "yandere"
                        creator = _data['creator']
                        if type(creator) == list:
                            creator = (lambda x: ", ".join(x))(creator)
                        characters = _data['characters']
                        source = _data['source']
                        found_json['data'] = {
                            "creator": creator, "characters": characters, "source": source}

                    elif index_id == 13:
                        # 13 -> konachan
                        # ext_urls, konachan_id, creator, material, characters, source
                        found_json['index'] = "konachan"
                        creator = _data['creator']
                        if type(creator) == list:
                            creator = (lambda x: ", ".join(x))(creator)
                        characters = _data['characters']
                        source = _data['source']
                        found_json['data'] = {
                            "creator": creator, "characters": characters, "source": source}
                    elif index_id == 18:
                        # 18-> H-Misc nhentai
                        found_json['index'] = "H-Misc"
                        source = _data['source']
                        creator = _data['creator']
                        if type(creator) == list:
                            creator = (lambda x: ", ".join(x))(creator)
                        found_json['data'] = {
                            "source": source, "creator": creator}
                        if (x := _data['jp_name']):
                            found_json['data']['jp_name'] = x
                        if (x := _data['eng_name']):
                            found_json['data']['eng_name'] = x
                    elif index_id == 31:
                        # 31 -> bcy.net
                        found_json['index'] = "bcy"
                        url = _data['ext_urls'][0]
                        title = _data['title']
                        member_name = _data['member_name']
                        found_json['data'] = {
                            'title': title, 'url': url, 'member_name': member_name}
                    elif index_id == 34:
                        # 34->da
                        found_json['index'] = 'da'
                        illust_id = _data['da_id']
                        found_json['data'] = {"illust_id": illust_id}
                    elif index_id == 38:
                        # 38 -> H-Misc (E-Hentai)
                        found_json['index'] = "H-Misc"
                        source = _data['source']
                        creator = _data['creator']
                        if type(creator) == list:
                            creator = (lambda x: ", ".join(x))(creator)
                        jp_name = _data['jp_name']
                        found_json['data'] = {
                            "source": source, "creator": creator, "jp_name": jp_name}
                    elif index_id == 39:
                        # 39 -> Artstation
                        found_json['index'] = "artstation"
                        url = _data['ext_urls'][0]
                        title = _data['title']
                        author_name = _data['author_name']
                        found_json['data'] = {
                            'title': title, 'url': url, 'author_name': author_name}
                    elif index_id == 41:
                        # 41->twitter
                        found_json['index'] = "twitter"
                        url = ",".join(_data['ext_urls'])
                        date = _data['created_at']
                        creator = _data['twitter_user_handle']
                        found_json['data'] = {"url": url,
                                              "date": date, "creator": creator}

                    else:
                        return Response(ACTION_FAILED, message=f"Unhandled Index {index_id}, check log for more infomation.")
                    #found_json['index'] = index
                    found_json['rate'] = rate
                    return Response(ACTION_SUCCESS, content=found_json)

                else:
                    return Response(ACTION_WARNING, message=f"rate: {rate}\nnot found... ;_;")

            else:
                # could potentially be negative
                if int(results['header']['long_remaining']) < 1:
                    return Response(ACTION_FAILED, message="Out of searches today. ")

                if int(results['header']['short_remaining']) < 1:
                    return Response(ACTION_FAILED, message="Out of searches in 30s. ")



# -*- coding:utf-8 -*-
import httpx
from bs4 import BeautifulSoup
from .response import *

class SingleRes():
    def __init__(self, title=None, title_url=None, author=None, author_url=None,thumb_url=None):
        self.title = title
        self.title_url = title_url
        self.author = author
        self.author_url = author_url
        self.thumbnail_url = thumb_url
        self.thumbnail = None


class Ascii2D:
    """ # Ascii2D search moudle #"""
    __instance = None

    def __init__(self,proxy=None):
        self.proxy = proxy

    def __new__(cls, *a, **k):
        if not cls.__instance:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def parse_html(self, data: httpx.Response) -> "list":
        soup = BeautifulSoup(data.text, 'html.parser')
        results = []
        for img in soup.find_all('div', attrs={'class': 'row item-box'}):
            img_url = "https://ascii2d.net" + str(img.img['src'])
            the_list = img.find_all('a')
            title = str(the_list[0].get_text())
            if title == "色合検索":
                results.append(SingleRes())
                continue
            title_url = str(the_list[0]["href"])
            author = str(the_list[1].get_text())
            author_url = str(the_list[1]["href"])
            results.append(SingleRes(title, title_url, author, author_url,img_url))
        return results

    async def search(self, url: str) -> "BaseResponse":
        try:
            client = httpx.AsyncClient(proxies=self.proxy,follow_redirects=True)
            # color_res = await client.post("https://ascii2d.net/search/multi", files=files)
            
            # 既然ascii2d能传图片url, 那就直接丢给它好了！
            color_response = await client.get("https://ascii2d.net/search/url/"+url,follow_redirects=True)
            #print(color_response.url)
            bovw_url = color_response.url.__str__().replace("/color/", "/bovw/")
            bovw_response = await client.get(bovw_url)
            color_results = self.parse_html(color_response)
            bovw_results = self.parse_html(bovw_response)
            color_results[1].thumbnail = (await client.get(color_results[1].thumbnail_url)).content
            bovw_results[1].thumbnail = (await client.get(bovw_results[1].thumbnail_url)).content
        #res = requests.post(ASCII2DURL, headers=headers, data=m, verify=False, **self.requests_kwargs)
            await client.aclose()
        except httpx.ReadTimeout:
            return BaseResponse(ACTION_FAILED,message="链接超时, 请检查网络是否通畅")
        except httpx.ProxyError:
            return BaseResponse(ACTION_FAILED,message="连接代理服务器出现错误, 请检查代理设置")

        if color_results[0].title:
            single = color_results[0]
            # 处理逻辑： 先看第一个返回结果是否带上title，如果有说明这张图已经被搜索过了，有直接结果
            # 如果第一个结果的title为空，那么直接返回第二个结果，带上缩略图让用户自行比对是否一致
            return BaseResponse(ACTION_SUCCESS, "get direct result from ascii2d color", {'index': "ascii2d颜色检索", 'title': single.title, 'url': single.title_url})
        elif bovw_results[0].title:
            single = bovw_results[0]
            return BaseResponse(ACTION_SUCCESS, "get direct result from ascii2d bovw", {'index': "ascii2d特征检索", 'title': single.title, 'url': single.title_url})
        else:

            return BaseResponse(ACTION_WARNING, "get possible results from ascii2d", [
                {'[ ascii2d': " 颜色检索 ]",  'title': color_results[1].title, 'title_url': color_results[1].title_url,
                "author":color_results[1].author,"author_url":color_results[1].author_url}, color_results[1].thumbnail,
                
                {'[ ascii2d': " 特征检索 ]",  'title': bovw_results[1].title, 'url': bovw_results[1].title_url,
                "author":bovw_results[1].author,"author_url":bovw_results[1].author_url}, bovw_results[1].thumbnail])



from typing import Any


ACTION_SUCCESS = 200
ACTION_WARNING = 300
ACTION_FAILED = 400

class BaseResponse:
    """AQUABOT 返回类

    * `` 当请求正常, 返回 ACTION_SUCCESS message , content`` 
    * `` 当请求异常, 返回 ACTION_WARNING 和message``
    * `` 当请求失败, 返回 ACTION_FAILED 和message``

    * ``status_code: int`` : 状态码
    * ``message: str`` : 信息
    * ``content: any`` : 返回内容
    """

    def __init__(self, status_code:int, message:str=None, content:Any=None,logger=None) -> None:
        self.status_code = status_code
        self.message = message
        self.content = content
        self.log = f"{self.status_code} | {self.message}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_imgsearch']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.1,<5.0.0',
 'httpx[http2]>=0.20.0,<1.0.0',
 'loguru>=0.5.1,<0.6.0']

setup_kwargs = {
    'name': 'nonebot-plugin-imgsearch',
    'version': '0.1.2',
    'description': 'A image search plugin for nonebot2',
    'long_description': '<p align="center">\n  <a href="https://v2.nonebot.dev/"><img src="https://raw.githubusercontent.com/nonebot/nonebot2/master/docs/.vuepress/public/logo.png" width="200" height="200" alt="nonebot"></a>\n</p>\n\n<div align="center">\n\n# nonebot-plugin-imgsearch\n\n_✨ NoneBot2 图片搜索插件 ✨_\n\n</div>\n\n<p align="center">\n  <a href="https://raw.githubusercontent.com/nonebot/nonebot2/master/LICENSE">\n    <img src="https://img.shields.io/github/license/nonebot/nonebot2.svg" alt="license">\n  </a>\n  <a href="https://pypi.python.org/pypi/nonebot-plugin-imgsearch">\n    <img src="https://img.shields.io/pypi/v/nonebot-plugin-imgsearch.svg" alt="pypi">\n  </a>\n  <img src="https://img.shields.io/badge/python-3.7+-blue.svg" alt="python">\n</p>\n\n## 快速启动\n\n* ### 1. 通过pip安装\n  ```\n  pip install nonebot-plugin-imgsearch\n  ```\n\n\n* ### 2. 申请 SAUCENAO API KEY\n  [点此申请](https://saucenao.com/user.php?page=search-api)  \n  `SauceNAO API KEY` 用于直接向saucenao请求搜索，无需模拟抓包，更快，更方便。  \n  一个 `API KEY` 每30s可请求5次，每24h可请求200次。\n\n* ### 3. 配置`.env`文件\n  在你的`.env.*`文件中添加如下内容：\n  ```\n  search_proxy: <your_proxy>\n  # 代理选项，可留空，如果你的bot位于国内，请填一个代理，\n  # 通常它长这样 -> search_proxy="http://127.0.0.1:7890"\n  \n  saucenao_api_key: <your_saucenao_api_key>\n  # 填你申请到的 SauceNAO API KEY, 是必选项。\n  ```\n\n\n\n## 使用教程\n\n  在群组或私聊中发送 `/search [图片]` 即可进行搜索，支持多张  \n  更多权限设置请修改`source code`\n',
    'author': 'bakashigure',
    'author_email': 'bakashigure@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bakashigure/nonebot_plugin_imgsearch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.3,<4.0.0',
}


setup(**setup_kwargs)
